"""
This module contains classes and functions involved in the
molecular conformer creation.
"""

import os
import glob
import numpy as np
import networkx as nx
import peleffy
from peleffy.utils import Logger
from peleffy.template import impact
from peleffy.topology import molecule
from peleffy.utils.toolkits import RDKitToolkitWrapper


class BCEConformations(object):
    """
    A class to produce a library of conformations from the output
    of the BCE server.
    """
    def __init__(self, topology_obj, bce_output_path):
        """
        Initializes a BCEConformations object.

        Parameters
        ----------
        topology_obj : An peleffy.topology.Topology
            A Topology object that contains the ligand's information
        bce_output_path: str
            Path where the output from the BCE server is stored
        """
        self._topology = topology_obj
        self._molecule = topology_obj.molecule
        self.bce_path = bce_output_path
        self.conformations_library = {}

    def calculate(self):
        """
        Calculate conformation library from the BCE output.
        """
        logger = Logger()
        logger.info(' - Calculating conformation library')
        self._calculate_all_conformations()

    def _calculate_all_conformations(self):
        clusters = sorted(glob.glob(os.path.join(self.bce_path,
                                                 "CLUSTERS", "CL*",
                                                 "cluster*.min.imaged.pdb")))
        if not clusters:
            raise ValueError("Path to the BCE output does not contain a "
                             + "CLUSTERS folder, please check if the path "
                             + "is correct.")
        clusters_order = self.order_clusters_min_distances(clusters)
        ordered_clusters = [clusters[x] for x in clusters_order]
        for cluster in ordered_clusters:
            self.calculate_cluster_offsets(cluster)


    def calculate_cluster_offsets(self, cluster_pdb):
        """
        Calculate dihedral angles from PDB.

        Parameters
        ----------
        cluster_pdb: str
            Path to the cluster representative conformation
        """
        rdkit_wrapper = RDKitToolkitWrapper()
        conformation_offsets = []
        # Use the input molecule as template since the cluster structures
        # probably will not have proper stereochemistry
        mol = molecule.Molecule(
            cluster_pdb, connectivity_template=self._molecule.rdkit_molecule)
        cluster_coordinates = mol.get_conformer()
        topology_to_cluster = {
            i: x for i, x
            in enumerate(rdkit_wrapper.get_substruct_match(mol,
                                                           self._molecule))}
        cluster_to_topology = {v: k for k, v in topology_to_cluster.items()}
        template = impact.Impact(self._topology)
        root_coordinates = cluster_coordinates[
            topology_to_cluster[find_index_root(template.topology,
                                                self._topology)]]
        for i, coords in enumerate(cluster_coordinates):
            conformation_offsets.append(
                [self._topology.atoms[cluster_to_topology[i]].PDB_name]
                + (coords-root_coordinates).tolist())
        self.conformations_library[cluster_pdb] = conformation_offsets

    def save(self, output_path):
        """
        Save the conformation library.

        Parameters
        ----------
        output_path: str
            Where to save the conformation library
        """
        with open(output_path, "w") as fw:
            fw.write("* CONFORMATIONS LIBRARY FILE\n")
            fw.write('* File generated with peleffy-{}\n'.format(
                peleffy.__version__))
            for conformation_file, conformation_values in \
                    self.conformations_library.items():
                fw.write("* File: {:s}\n".format(conformation_file))
                fw.write("{:s} {:d} {:d}\n".format(
                    self._molecule.tag.upper(),
                    len(conformation_values),
                    len(self.conformations_library)))
                for values in conformation_values:
                    fw.write("{:s} {:5f} {:5f} {:5f}\n".format(*values))
                fw.write("ENDCONFORMATION\n")
            fw.write("END\n")

    def order_clusters_min_distances(self, clusters):
        """
        Order a list of structures in a way that jumps to the next structure
        involves the minimum change.

        Parameters
        ----------
        clusters: list
            Path to the structures

        Returns
        -------
        min_path : list
            Nodes on the graph that minimise heuristically the total distance
        """
        rdkit_wrapper = RDKitToolkitWrapper()
        distances = np.zeros((len(clusters), len(clusters)))
        cluster_molecules = []
        for cluster in clusters:
            # use the input molecule as template since the cluster structures
            # probably will not have proper stereochemistry
            cluster_molecules.append(molecule.Molecule(
                cluster,
                connectivity_template=self._molecule.rdkit_molecule))
        for i, cluster_mol in enumerate(cluster_molecules):
            for j, cluster_mol_2 in enumerate(cluster_molecules[i+1:],
                                              start=i+1):
                rmsd_value = rdkit_wrapper.get_rmsd(cluster_mol, cluster_mol_2)
                distances[i, j] = rmsd_value
                distances[j, i] = rmsd_value

        min_path = find_optimal_path_from_matrix(distances)

        return min_path


def find_optimal_path_from_matrix(distances):
    """
    Find a Minimum Spanning Tree from a distance matrix. It uses networkx to
    create the graph and the MST

    Parameters
    ----------
    distances : a numpy.Array object
        The numpy array of shape (N, N), where N is the number of
        clustered molecules. It contains the RMSD value for each pair
        of clusters

    Returns
    -------
    min_path : list
        Nodes on the graph that minimize heuristically the total distance
    """
    graph = nx.from_numpy_matrix(distances)

    min_dist = 1e6
    min_path = None

    for node in graph:
        dist, path = find_heuristic_path(graph.copy(), distances, node)
        if dist < min_dist:
            min_dist = dist
            min_path = path

    return min_path

def find_heuristic_path(graph, distances, start_node):
    """
    Given a graph, the corresponding distances matrix and a starting node,
    it finds the heuristic path that minimizes the total distance.

    Parameters
    ----------
    graph : a networkx.Graph object
        The graph containing clusters as nodes and labelled edges
        according to relative RMSD distances between pairs
    distances : a numpy.Array object
        The numpy array of shape (N, N), where N is the number of
        clustered molecules. It contains the RMSD value for each pair
        of clusters
    start_node : int
        The id of the node from which the path will be computed

    Returns
    -------
    dist : float
        The total distance of the final path
    path : list[int]
        The list of node ids that conform the final path
    """
    n = distances.shape[0]
    path = [start_node]
    dist = 0

    while len(path) < n:
        node = path[-1]
        new_neighbor = min(graph[node].items(), key=lambda x: x[1]["weight"])
        graph.remove_node(node)
        path.append(new_neighbor[0])
        dist += new_neighbor[1]['weight']

    dist += distances[path[-1], start_node]

    return dist, path


def find_index_root(sorted_topology, topology):
    """
    It finds the index of the root atom in the original topology
    that matches with the one from the topology that fullfills the Impact
    template rules.

    Parameters
    ----------
    sorted_topology : a peleffy.topology.Topology object
        The topology sorted to fullfill the Impact template format
    topology : a peleffy.topology.Topology object
        The original topology object

    Returns
    -------
    index_root : int
        The index of the root atom in the original topology
    """
    atom_name = sorted_topology.atoms[0].PDB_name

    for i, atom in enumerate(topology.atoms):
        if atom.PDB_name == atom_name:
            return i

    return None



"""
This module contains the tests to check all available force fields in
peleffy.
"""


class TestOpenForceField(object):
    """
    It wraps all tests that check the OpenForceField class.
    """

    FORCE_FIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

    def test_name(self):
        """It checks the name assignment."""

        from peleffy.forcefield import OpenForceField

        openff = OpenForceField(self.FORCE_FIELD_NAME)

        assert openff.name == self.FORCE_FIELD_NAME, \
            'Unexpected force field name'

    def test_type(self):
        """It checks the type assignment."""

        from peleffy.forcefield import OpenForceField

        openff = OpenForceField(self.FORCE_FIELD_NAME)

        assert openff.type == 'OpenFF', \
            'Unexpected force field type'

    def test_charge_calculator_selector(self):
        """It checks the charge calculator selector."""
        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField
        import peleffy

        dummy_mol = Molecule()

        # Check default selection
        openff = OpenForceField(self.FORCE_FIELD_NAME)
        calculator = openff._get_charge_calculator(None, dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.Am1bccCalculator), \
            'Invalid default charge calculator: ' \
            + '{}'.format(type(calculator))

        # Check custom selection 1
        openff = OpenForceField(self.FORCE_FIELD_NAME)
        calculator = openff._get_charge_calculator('gasteiger', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.GasteigerCalculator), \
            'Invalid custom selection 1 for the charge calculator'

        # Check custom selection 1
        openff = OpenForceField(self.FORCE_FIELD_NAME)
        calculator = openff._get_charge_calculator('opls2005', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.OPLSChargeCalculator), \
            'Invalid custom selection 2 for the charge calculator'

    def test_parameterizer(self):
        """It checks the parameterized method."""

        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField
        from peleffy.utils import (get_data_file_path,
                                   convert_all_quantities_to_string)
        from .utils import compare_dicts
        import json

        # Load molecule 1
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        openff = OpenForceField(self.FORCE_FIELD_NAME)

        # Obtain force field parameters
        parameters = openff.parameterize(molecule)

        writable_parameters = convert_all_quantities_to_string(parameters)

        reference_file = get_data_file_path(
            'tests/MET_openff-1.2.1_parameters.json')

        with open(reference_file) as f:
            compare_dicts(writable_parameters, json.load(f))

        # Load molecule
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        openff = OpenForceField(self.FORCE_FIELD_NAME)

        # Obtain force field parameters
        parameters = openff.parameterize(molecule)

        writable_parameters = convert_all_quantities_to_string(parameters)

        reference_file = get_data_file_path(
            'tests/ETL_openff-1.2.1_parameters.json')

        with open(reference_file) as f:
            compare_dicts(writable_parameters, json.load(f))

    def test_OPLS_charge_method_with_openff(self):
        """
        It tests the obtaining of OPLS2005 charges using the OpenFF
        force field.
        """
        import peleffy
        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenForceField
        from peleffy.utils import get_data_file_path

        # Load molecule
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))

        # Load OpenFF force field with the OPLS2005 charge method
        openff = OpenForceField(self.FORCE_FIELD_NAME)

        charge_calculator = openff._get_charge_calculator(
            charge_method='OPLS2005',
            molecule=molecule)

        assert isinstance(
            charge_calculator,
            peleffy.forcefield.calculators.OPLSChargeCalculator), \
            'Unexpected charge calculator method'


class TestOPLS2005ForceField(object):
    """
    It wraps all tests that check the OPLS2005ForceField class.
    """

    FORCE_FIELD_NAME = 'OPLS2005'

    def test_name(self):
        """It checks the name assignment."""

        from peleffy.forcefield import OPLS2005ForceField

        oplsff = OPLS2005ForceField()

        assert oplsff.name == self.FORCE_FIELD_NAME, \
            'Unexpected force field name'

    def test_type(self):
        """It checks the type assignment."""

        from peleffy.forcefield import OPLS2005ForceField

        oplsff = OPLS2005ForceField()

        assert oplsff.type == 'OPLS2005', \
            'Unexpected force field type'

    def test_charge_calculator_selector(self):
        """It checks the charge calculator selector."""
        from peleffy.topology import Molecule
        from peleffy.forcefield import OPLS2005ForceField
        import peleffy

        dummy_mol = Molecule()

        # Check default selection
        oplsff = OPLS2005ForceField()
        calculator = oplsff._get_charge_calculator(None, dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.OPLSChargeCalculator), \
            'Invalid default charge calculator: ' \
            + '{}'.format(type(calculator))

        # Check custom selection 1
        oplsff = OPLS2005ForceField()
        calculator = oplsff._get_charge_calculator('gasteiger', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.GasteigerCalculator), \
            'Invalid custom selection 1 for the charge calculator'

        # Check custom selection 1
        oplsff = OPLS2005ForceField()
        calculator = oplsff._get_charge_calculator('am1bcc', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.Am1bccCalculator), \
            'Invalid custom selection 2 for the charge calculator'

    def test_parameterizer(self):
        """It checks the parameterized method."""

        from peleffy.topology import Molecule
        from peleffy.forcefield import OPLS2005ForceField
        from peleffy.utils import (get_data_file_path,
                                   convert_all_quantities_to_string)
        from .utils import compare_dicts, parameterize_opls2005
        import json

        # Load molecule 1
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        oplsff = OPLS2005ForceField()
        ffld_file = get_data_file_path('tests/MET_ffld_output.txt')

        parameters = parameterize_opls2005(oplsff, molecule, ffld_file)

        writable_parameters = convert_all_quantities_to_string(parameters)

        reference_file = get_data_file_path(
            'tests/MET_opls2005_parameters.json')

        with open(reference_file) as f:
            compare_dicts(writable_parameters, json.load(f))

        # Load molecule 1
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        oplsff = OPLS2005ForceField()
        ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')

        parameters = parameterize_opls2005(oplsff, molecule, ffld_file)

        writable_parameters = convert_all_quantities_to_string(parameters)

        reference_file = get_data_file_path(
            'tests/ETL_opls2005_parameters.json')

        with open(reference_file) as f:
            compare_dicts(writable_parameters, json.load(f))


class TestOpenFFOPLS2005ForceField(object):
    """
    It wraps all tests that check the OpenFFOPLS2005ForceField class.
    """

    FORCE_FIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

    def test_name(self):
        """It checks the name assignment."""

        from peleffy.forcefield import OpenFFOPLS2005ForceField

        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)

        assert hybridff.name == self.FORCE_FIELD_NAME + ' + OPLS2005', \
            'Unexpected force field name'

    def test_type(self):
        """It checks the type assignment."""

        from peleffy.forcefield import OpenFFOPLS2005ForceField

        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)

        assert hybridff.type == 'OpenFF + OPLS2005', \
            'Unexpected force field type'

    def test_charge_calculator_selector(self):
        """It checks the charge calculator selector."""
        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenFFOPLS2005ForceField
        import peleffy

        dummy_mol = Molecule()

        # Check default selection
        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)
        calculator = hybridff._get_charge_calculator(None, dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.Am1bccCalculator), \
            'Invalid default charge calculator: ' \
            + '{}'.format(type(calculator))

        # Check custom selection 1
        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)
        calculator = hybridff._get_charge_calculator('gasteiger', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.GasteigerCalculator), \
            'Invalid custom selection 1 for the charge calculator'

        # Check custom selection 1
        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)
        calculator = hybridff._get_charge_calculator('opls2005', dummy_mol)

        assert isinstance(
            calculator,
            peleffy.forcefield.calculators.OPLSChargeCalculator), \
            'Invalid custom selection 2 for the charge calculator'

    def test_parameterizer(self):
        """It checks the parameterized method."""

        def check(hybridff, molecule, ffld_file, reference_file):
            """
            It checks the parameters obtained by the force field.

            Parameters
            ----------
            hybridff : an OpenFFOPLS2005ForceField object
                The hybrid force field to employ in the parameterization
                along with the ffld_file
            molecule : a peleffy.topology.Molecule
                The peleffy's Molecule object to parameterize with the
                ffld file
            ffld_file : str
                The path to the precomputed ffld file from where the
                parameters will be extracted
            reference_file : str
                The path to the file containing the reference parameters
            """
            parameters = parameterize_openffopls2005(hybridff,
                                                     molecule,
                                                     ffld_file)

            writable_parameters = convert_all_quantities_to_string(parameters)

            with open(reference_file) as f:
                compare_dicts(writable_parameters, json.load(f))

        from peleffy.topology import Molecule
        from peleffy.forcefield import OpenFFOPLS2005ForceField
        from peleffy.utils import (get_data_file_path,
                                   convert_all_quantities_to_string)
        from .utils import compare_dicts, parameterize_openffopls2005
        import json

        # Load molecule 1
        molecule = Molecule(get_data_file_path('ligands/methane.pdb'))
        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)
        ffld_file = get_data_file_path('tests/MET_ffld_output.txt')

        # 1st check
        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/MET_openff-1.2.1_opls2005_parameters1.json'))

        # 2nd check
        hybridff.set_nonbonding_parameters('OPLS2005')

        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/MET_openff-1.2.1_opls2005_parameters2.json'))

        # 3rd check
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_bond_parameters('OPLS2005')

        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/MET_openff-1.2.1_opls2005_parameters3.json'))

        # 4th check
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_bond_parameters('OpenFF')
        hybridff.set_angle_parameters('OPLS2005')

        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/MET_openff-1.2.1_opls2005_parameters4.json'))

        # 5th check
        hybridff.set_nonbonding_parameters('OPLS2005')
        hybridff.set_bond_parameters('OPLS2005')
        hybridff.set_angle_parameters('OPLS2005')

        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/MET_openff-1.2.1_opls2005_parameters5.json'))

        # Load molecule 2
        molecule = Molecule(get_data_file_path('ligands/ethylene.pdb'))
        hybridff = OpenFFOPLS2005ForceField(self.FORCE_FIELD_NAME)
        ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')

        # 1st check
        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/ETL_openff-1.2.1_opls2005_parameters1.json'))

        # 2nd check
        hybridff.set_nonbonding_parameters('OpenFF')
        hybridff.set_dihedral_parameters('OPLS2005')

        check(hybridff, molecule, ffld_file,
              get_data_file_path(
                  'tests/ETL_openff-1.2.1_opls2005_parameters2.json'))


class TestForceFieldSelector(object):
    """
    It wraps all tests that check the force field selector class.
    """

    def test_get_by_name(self):
        """It checks the get_by_name method."""

        from peleffy.forcefield import ForceFieldSelector
        from peleffy.forcefield import OpenForceField
        from peleffy.forcefield import OPLS2005ForceField

        selector = ForceFieldSelector()

        forcefield = selector.get_by_name('openff_unconstrained-1.0.0.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.0.1.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.1.0.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.1.1.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.2.0.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.2.1.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('openff_unconstrained-1.3.0.offxml')

        assert isinstance(forcefield, OpenForceField), \
            'Unexpected force field type'

        forcefield = selector.get_by_name('OPLS2005')

        assert isinstance(forcefield, OPLS2005ForceField), \
            'Unexpected force field type'



"""
This module contains the tests to check the peleffy's Impact template builder.
"""

import pytest

import tempfile

from peleffy.utils import get_data_file_path, temporary_cd
from peleffy.topology import Molecule
from peleffy.template import Impact
from peleffy.tests.utils import compare_files
from peleffy.topology import Topology
from peleffy.forcefield import OpenForceField, OPLS2005ForceField


class TestImpactTemplate(object):
    """It tests the Impact class that writes a PELE's Impact template"""

    OPENFF_FORCEFIELD = 'openff_unconstrained-1.2.1.offxml'

    def test_input(self):
        """
        It tests that the topology given to Impact() is of the correct
        format, peleffy.topology.Topology.
        """
        from peleffy.forcefield.parameters import BaseParameterWrapper

        LIGAND_PATH = 'ligands/benzene.pdb'

        ligand_path = get_data_file_path(LIGAND_PATH)
        molecule = Molecule(ligand_path)

        parameters = BaseParameterWrapper()

        topology = Topology(molecule, parameters)

        # Impact() gets nothing as argument
        with pytest.raises(TypeError):
            _ = Impact()

        # Impact() gets a non Topology object
        with pytest.raises(TypeError):
            _ = Impact('passing a str instead of a Topology')

        # This should work
        _ = Impact(topology)

    def test_writer_OFF(self):
        """
        It tests the writer attribute of the Impact class using OFF
        to parameterize.
        """

        TEMPLATE_METZ = get_data_file_path('tests/metz')
        TEMPLATE_MATZ = get_data_file_path('tests/malz')
        TEMPLATE_ETLZ = get_data_file_path('tests/etlz')

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                # Generates the template for methane
                pdb_path = get_data_file_path('ligands/methane.pdb')
                molecule = Molecule(pdb_path)
                openff = OpenForceField(self.OPENFF_FORCEFIELD)
                parameters = openff.parameterize(molecule)
                topology = Topology(molecule, parameters)

                # Generates the impact template for methane
                impact = Impact(topology)
                impact.to_file('metz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_METZ, file2='metz')

                # Generates the template for malonate
                pdb_path = get_data_file_path('ligands/malonate.pdb')
                molecule = Molecule(pdb_path)
                openff = OpenForceField(self.OPENFF_FORCEFIELD)
                parameters = openff.parameterize(molecule)
                topology = Topology(molecule, parameters)

                # Generates the impact template for malonate
                impact = Impact(topology)
                impact.to_file('malz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_MATZ, file2='malz')

                # Generates the template for ethylene
                pdb_path = get_data_file_path('ligands/ethylene.pdb')
                molecule = Molecule(pdb_path, tag='ETL')  # Note that in this case we are assigning a tag to the molecule which will be used in the Impact template
                openff = OpenForceField(self.OPENFF_FORCEFIELD)
                parameters = openff.parameterize(molecule)
                topology = Topology(molecule, parameters)

                # Generates the impact template for ethylene
                impact = Impact(topology)
                impact.to_file('etlz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_ETLZ, file2='etlz')

    def test_writer_OPLS(self):
        """
        It tests the writer attribute of the Impact class using OPLS to parameterize.
        """
        from .utils import parameterize_opls2005

        TEMPLATE_METZ_OPLS = get_data_file_path('tests/OPLS_metz')
        TEMPLATE_MALZ_OPLS = get_data_file_path('tests/OPLS_malz')
        TEMPLATE_ETLZ_OPLS = get_data_file_path('tests/OPLS_etlz')

        with tempfile.TemporaryDirectory() as tmpdir:
            with temporary_cd(tmpdir):
                # Generates the template for methane using OPLS
                opls2005 = OPLS2005ForceField()
                pdb_path = get_data_file_path('ligands/methane.pdb')
                molecule = Molecule(pdb_path)
                ffld_file = get_data_file_path('tests/MET_ffld_output.txt')
                parameters = parameterize_opls2005(opls2005,
                                                   molecule,
                                                   ffld_file)
                topology = Topology(molecule, parameters)

                # Generates the impact template for methane
                impact = Impact(topology)
                impact.to_file('metz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_METZ_OPLS, file2='metz')

                # Generates the template for malonate using OPLS
                opls2005 = OPLS2005ForceField()
                pdb_path = get_data_file_path('ligands/malonate.pdb')
                molecule = Molecule(pdb_path)
                ffld_file = get_data_file_path('tests/MAL_ffld_output.txt')
                parameters = parameterize_opls2005(opls2005,
                                                   molecule,
                                                   ffld_file)
                topology = Topology(molecule, parameters)

                # Generates the impact template for malonate
                impact = Impact(topology)
                impact.to_file('malz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_MALZ_OPLS, file2='malz')

                # Generates the template for ethylene using OPLS
                opls2005 = OPLS2005ForceField()
                pdb_path = get_data_file_path('ligands/ethylene.pdb')
                molecule = Molecule(pdb_path, tag='ETL')
                ffld_file = get_data_file_path('tests/ETL_ffld_output.txt')
                parameters = parameterize_opls2005(opls2005,
                                                   molecule,
                                                   ffld_file)
                topology = Topology(molecule, parameters)

                # Generates the impact template for ethylene
                impact = Impact(topology)
                impact.to_file('etlz')

                # Compare the reference template and the generated template
                compare_files(file1=TEMPLATE_ETLZ_OPLS, file2='etlz')

    def test_get_absolute_parent_atom(self):
        """
        It tests the _get_absolute_parent_atom method used in the building
        process of the Impact template.
        """

        LIGAND_PATH = get_data_file_path('ligands/malonate.pdb')
        FORCEFIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

        molecule = Molecule(LIGAND_PATH)

        openff = OpenForceField(FORCEFIELD_NAME)

        parameters = openff.parameterize(molecule, charge_method='dummy')

        topology = Topology(molecule, parameters)

        impact = Impact(topology)

        absolute_parent = impact._get_absolute_parent_atom()

        assert absolute_parent.PDB_name == '_C2_', \
            'Unexpected absolute parent atom: {}'.format(absolute_parent)

    def test_get_all_childs_of_atom(self):
        """
        It tests the _get_all_childs_of_atom method used in the building
        process of the Impact template.
        """

        LIGAND_PATH = get_data_file_path('ligands/malonate.pdb')
        FORCEFIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

        molecule = Molecule(LIGAND_PATH)

        openff = OpenForceField(FORCEFIELD_NAME)

        parameters = openff.parameterize(molecule, charge_method='dummy')

        topology = Topology(molecule, parameters)

        impact = Impact(topology)

        absolute_parent = impact._get_absolute_parent_atom()

        childs = impact._get_all_childs_of_atom(absolute_parent, 'side chain')

        assert [a.PDB_name for a in childs] == \
            ['_C1_', '_C3_'], \
            'Unexpected side-chain-child atoms: {}'.format(childs)

        childs = impact._get_all_childs_of_atom(absolute_parent, 'core')

        assert [a.PDB_name for a in childs] == \
            ['_H1_', '_H2_'], \
            'Unexpected core-child atoms: {}'.format(childs)

    def test_get_core_atoms(self):
        """
        It tests the _get_core_atoms method used in the building
        process of the Impact template.
        """

        LIGAND_PATH = get_data_file_path('ligands/malonate.pdb')
        FORCEFIELD_NAME = 'openff_unconstrained-1.2.1.offxml'

        molecule = Molecule(LIGAND_PATH)

        openff = OpenForceField(FORCEFIELD_NAME)

        parameters = openff.parameterize(molecule, charge_method='dummy')

        topology = Topology(molecule, parameters)

        impact = Impact(topology)

        core_atoms = impact._get_core_atoms()

        assert [a.PDB_name for a in core_atoms] == \
            ['_C2_', '_H1_', '_H2_'], \
            'Unexpected core atoms: {}'.format(core_atoms)



"""
This module contains classes and functions involved in the manipulation of
PELE's solvent templates.
"""

from simtk import unit

from peleffy.utils import get_data_file_path
from peleffy.utils import Logger


class _SolventWrapper(object):
    """
    A wrapper for any solvent-like class.
    """
    _ff_file = None
    _name = None
    _compatibility = None

    def __init__(self, topologies):
        """
        Initializes a SolventWrapper object.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template
        """
        self._topologies = topologies

        self._initialize_dicts()

    def _initialize_dicts(self):
        """
        It initializes empty radii and scales dictionaries. It also handles the
        possibility of dealing with a single topology or multiple of topologies.
        """

        _multiple_topologies = isinstance(self.topologies, list)

        if not _multiple_topologies:
            self._topologies = [self._topologies]

        self._radii = \
            [dict.fromkeys([tuple((idx, )) for idx in
                            range(0, len(topology.atoms))], unit.Quantity())
             for topology in self._topologies]
        self._scales = \
            [dict.fromkeys([tuple((idx, )) for idx in
                            range(0, len(topology.atoms))], unit.Quantity())
             for topology in self._topologies]

    @property
    def name(self):
        """
        The name of the solvent.

        Returns
        -------
        name : str
            The name of this solvent object.
        """
        return self._name

    @property
    def topologies(self):
        """
        The peleffy's Topology.

        Returns
        -------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template
        """
        return self._topologies

    @property
    def radii(self):
        """
        A list of the dicts of radii of the parameterized molecules.

        Returns
        -------
        radii : list[dict[atom indexes: simtk.unit.Quantity]]
            The radius assigned to each atom of the molecule
        """
        return self._radii

    @property
    def scales(self):
        """
        A list of the dicts of scales of the parameterized molecules.

        Returns
        -------
        scales : list[dict[atom indexes: simtk.unit.Quantity]]
            The scale assigned to each atom of the molecule
        """
        return self._scales


class _OpenFFCompatibleSolvent(_SolventWrapper):
    """
    Implementation of a solvent-template generator compatible with
    PELE's OpenFF implementation.
    """

    _compatibility = 'openff'

    def __init__(self, topologies):
        """
        It initializes an OpenFFCompatibleSolvent.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template
        """
        super().__init__(topologies)

        self._initialize_from_topology()

    def _initialize_from_topology(self):
        """
        Initializes a SolventWrapper object using a peleffy's
        molecular Topology.
        """
        logger = Logger()
        logger.info(' - Generating solvent parameters')

        from peleffy.utils.toolkits import OpenForceFieldToolkitWrapper

        off_toolkit = OpenForceFieldToolkitWrapper()
        GBSA_handler = off_toolkit.get_parameter_handler_from_forcefield(
            'GBSA', self._ff_file)

        self._solvent_dielectric = GBSA_handler.solvent_dielectric
        self._solute_dielectric = GBSA_handler.solute_dielectric
        self._surface_area_penalty = GBSA_handler.surface_area_penalty
        self._solvent_radius = GBSA_handler.solvent_radius

        from peleffy.forcefield import OpenForceField

        forcefield = OpenForceField(self._ff_file)

        for idx, topology in enumerate(self.topologies):
            parameters = forcefield.parameterize(topology.molecule,
                                                 charge_method='dummy')
            self._radii[idx] = parameters['GBSA_radii']
            self._scales[idx] = parameters['GBSA_scales']

    def to_dict(self):
        """
        Returns this OpenFFCompatibleSolvent object as a dictionary.

        Returns
        -------
        data : dict
            A dictionary containing the data of this SolventWrapper object
        """
        data = dict()
        data['SolventParameters'] = dict()
        data['SolventParameters']['Name'] = self.name
        data['SolventParameters']['General'] = dict()
        data['SolventParameters']['General']['solvent_dielectric'] = \
            round(self.solvent_dielectric, 5)
        data['SolventParameters']['General']['solute_dielectric'] = \
            round(self.solute_dielectric, 5)
        data['SolventParameters']['General']['solvent_radius'] = \
            round(self.solvent_radius.value_in_unit(unit.angstrom), 5)
        data['SolventParameters']['General']['surface_area_penalty'] = \
            round(self.surface_area_penalty.value_in_unit(
                unit.kilocalorie / (unit.angstrom**2 * unit.mole)), 8)

        for topology, radii, scales in zip(self._topologies,
                                           self._radii, self._scales):
            data['SolventParameters'][topology.molecule.tag] = dict()

            atom_names = topology.molecule.get_pdb_atom_names()

            for atom, name in zip(topology.molecule.rdkit_molecule.GetAtoms(),
                                  atom_names):
                name = name.replace(' ', '_')
                index = atom.GetIdx()
                data['SolventParameters'][topology.molecule.tag][name] = \
                    {'radius': round(radii[tuple((index, ))]
                                     .value_in_unit(unit.angstrom), 5),
                     'scale': round(scales[tuple((index, ))], 5)}
        return data

    def to_file(self, path):
        """
        Writes this OpenFFCompatibleSolvent object to a file.

        Parameters
        ----------
        path : str
            Path to save the output file to
        """
        import json
        with open(path, 'w') as file:
            json.dump(self.to_dict(), file, indent=4)

    @property
    def solvent_dielectric(self):
        """
        The solvent dielectric value of this solvent object.

        Returns
        -------
        solvent_dielectric : float
            The solvent dielectric value
        """
        return self._solvent_dielectric

    @property
    def solute_dielectric(self):
        """
        The solute dielectric value of this solvent object.

        Returns
        -------
        solute_dielectric : float
            The solute dielectric value
        """
        return self._solute_dielectric

    @property
    def surface_area_penalty(self):
        """
        The surface area penalty value of this solvent object.

        Returns
        -------
        surface_area_penalty : float
            The surface area penalty value
        """
        return self._surface_area_penalty

    @property
    def solvent_radius(self):
        """
        The solvent radius value of this solvent object.

        Returns
        -------
        solvent_radius : float
            The solvent radius value
        """
        return self._solvent_radius


class _OPLS2005CompatibleSolvent(_SolventWrapper):
    """
    Implementation of a solvent-template generator compatible with
    PELE's OPLS2005 implementation.
    """

    _compatibility = 'opls2005'
    _PARAMS_PATH = get_data_file_path('parameters/solventParamsHCTOBC.txt')

    def __init__(self, topologies):
        """
        It initializes an OPLS2005CompatibleSolvent.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template
        """
        super().__init__(topologies)

        self._initialize_from_topology()

    def _initialize_from_topology(self):
        """
        Initializes a SolventWrapper object using a peleffy's
        molecular Topology.
        """
        from peleffy.forcefield import OPLS2005ForceField
        from peleffy.forcefield.parameters import OPLS2005ParameterWrapper

        for idx, topology in enumerate(self.topologies):
            # Parameterize with OPLS2005 only if the parameters in topology
            # are not obtained with OPLS2005
            if isinstance(topology.parameters, OPLS2005ParameterWrapper):
                parameters = topology.parameters
            else:
                forcefield = OPLS2005ForceField()
                parameters = forcefield.parameterize(topology.molecule)

            self._radii[idx] = parameters['GBSA_radii']
            self._scales[idx] = parameters['GBSA_scales']

    def to_file(self, path):
        """
        Writes this OPLS2005CompatibleSolvent object to a file
        compatible with PELE.

        Parameters
        ----------
        path : str
            Path to save the output file to
        """

        # Load parameters for standard residues
        with open(self._PARAMS_PATH) as f:
            standard_params = f.read()

        with open(path, 'w') as f:
            # Write parameters for standard residues
            f.write(standard_params)

            # Write parameters for non standard residues
            for topology, radii, scales in zip(self.topologies,
                                               self.radii, self.scales):

                atom_names = [param.replace('_', '') for param in
                              topology.parameters['atom_names']]

                for atom_name, atom_type, scale, radi in \
                        zip(atom_names, topology.parameters['atom_types'],
                            scales, radii):

                    f.write(topology.molecule.tag + 'Z'.upper() + '   '
                            + atom_name + '   '
                            + atom_type + '    '
                            + str(scale) + '   '
                            + str(radi._value) + '\n')


class OBC1(_OpenFFCompatibleSolvent):
    """
    Implementation of the OBC1 solvent.
    """

    _ff_file = get_data_file_path('forcefields/GBSA_OBC1-1.0.offxml')
    _name = 'OBC1'

    def __init__(self, topologies):
        """
        Initializes an OBC1 object.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template
        """
        # Not implemented in PELE
        logger = Logger()
        logger.warning('OBC1 is not implemented in PELE')

        super().__init__(topologies)


class OBC2(_OpenFFCompatibleSolvent):
    """
    Implementation of the OBC2 solvent.
    """

    _ff_file = get_data_file_path('forcefields/GBSA_OBC2-1.0.offxml')
    _name = 'OBC2'

    def __init__(self, topologies):
        """
        Initializes an OBC2 object.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template

        Examples
        --------

        Generate the solvent parameters of a molecule

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule)

        >>> from peleffy.topology import Topology

        >>> topology = Topology(molecule, parameters)

        >>> from peleffy.solvent import OBC2

        >>> solvent = OBC2(topology)
        >>> solvent.to_file('OBC_parameters.txt')

        """
        super().__init__(topologies)


class OPLSOBC(_OPLS2005CompatibleSolvent):
    """
    It defines a template generator for OBC compatible with the OPLS2005
    force field implemented in PELE.
    """
    _name = 'OBC'

    def __init__(self, topologies):
        """
        Initializes an OPLSOBC object.

        Parameters
        ----------
        topologies : a Topology object or list[Topology object]
            The molecular topology representation to write as a
            Impact template

        Examples
        --------

        Generate the solvent parameters of a molecule

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OPLS2005ForceField

        >>> opls2005 = OPLS2005ForceField()
        >>> parameters = opls2005.parameterize(molecule)

        >>> from peleffy.topology import Topology

        >>> topology = Topology(molecule, parameters)

        >>> from peleffy.solvent import OPLSOBC

        >>> solvent = OPLSOBC(topology)
        >>> solvent.to_file('OBC_parameters.txt')

        """
        super().__init__(topologies)



"""
This module contains a set of classes and methods designed to handle
input data.
"""
from peleffy.utils import Logger


class PDBFile(object):
    """
    It handles an input PDB file and allows the extraction of multiple molecules
    as peleffy.topology.Molecule objects.
    """

    def __init__(self, path):
        """
        It initializes a PDB object through a PDB file.

        Parameters
        ----------
        path : str
            The path to the PDB with the molecules structures.

        Examples
        --------

        Load all the hetero molecules from a PDB

        >>> from peleffy.utils.input import PDBFile

        >>> PDBreader = PDBFile('/path/to/pdb.pdb')
        >>> molecules = PDBreader.get_hetero_molecules()

        Load from a PDB the hetero atom in the L chain.

        >>> from peleffy.utils.input import PDBFile

        >>> PDBreader = PDBFile('/path/to/pdb.pdb')
        >>> molecule  = PDBreader.get_molecule_from_chain(selected_chain = 'L')

        """
        with open(path, 'r') as pdb_file:
            self.pdb_content = pdb_file.readlines()

    def _extract_molecules_from_chain(self, chain, rotamer_resolution,
                                     exclude_terminal_rotamers,
                                     allow_undefined_stereo,
                                     core_constraints):
        """
        It extracts all hetero molecules found in the selected the chain
        of a PDB file.

        Parameters
        ----------
        chain_id : str
            Chain ID.
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False
        core_constraints : list[int or str]
            It defines the list of atoms to constrain in the core, thus,
            the core will be forced to contain them. Atoms can be specified
            through integers that match the atom index or strings that
            match with the atom PDB name

        Returns
        -------
        molecules : list[peleffy.topology.Molecule object]
            Selected molecules
        """
        from peleffy.topology.molecule import Molecule

        # Check if there is more than one hetero molecule in the same chain
        residues_ids = set([line[22:26].strip() for line in self.pdb_content
                            if line.startswith('HETATM')
                            and line[21:22] == chain
                            and not line[17:20].strip() == 'HOH'])

        molecules = []
        for residue_id in residues_ids:
            res_name = set([line[17:20].strip() for line in self.pdb_content
                            if line.startswith('HETATM')
                            and line[21:22] == chain
                            and line[22:26].strip() == residue_id])

            # Select which atoms compose this hetero molecule
            atom_ids = [line[6:11].strip() for line in self.pdb_content
                        if line.startswith('HETATM')
                        and line[21:22] == chain
                        and line[22:26].strip() == residue_id]

            # Extract the PDB block of the molecule
            pdb_block = [line for line in self.pdb_content
                         if (line.startswith('HETATM') or
                             line.startswith('CONECT'))
                         and any(' {} '.format(a) in line for a in atom_ids)]

            try:
                molecules.append(
                    Molecule(pdb_block=''.join(pdb_block),
                             rotamer_resolution=rotamer_resolution,
                             exclude_terminal_rotamers=exclude_terminal_rotamers,
                             allow_undefined_stereo=allow_undefined_stereo,
                             core_constraints=core_constraints))
            except Exception as e:
                log = Logger()
                log.warning(' - Skipping {} '.format(list(res_name)[0])
                            + 'from chain {}'.format(chain))
                log.warning('  - The following exception was raised: '
                            + '{}'.format(e))

        return molecules

    def get_hetero_molecules(self, rotamer_resolution=30,
                             exclude_terminal_rotamers=True,
                             allow_undefined_stereo=False,
                             ligand_core_constraints=[],
                             ligand_resname=None):
        """
        It returns a list of peleffy.topology.Molecule objects with all the
        hetero molecules contained in the PDB.

        Parameters
        ----------
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False
        ligand_core_constraints : list[int or str]
            It defines the list of atoms to constrain in the core of the
            ligand, thus, the core will be forced to contain them.
            Atoms can be specified through integers that match the atom index
            or strings that match with the atom PDB name
        ligand_resname : str
            Residue name of the ligand. Default is None

        Returns
        -------
        molecules : list[peleffy.topology.Molecule]
            List of the multiple molecules in the PDB file
        """
        chain_ids = set([line[21:22] for line in self.pdb_content
                         if line.startswith('HETATM')
                         and not line[17:20].strip() == 'HOH'])

        # Assign core constraints to the ligand, if specified
        d = {}
        for chain in chain_ids:
            resnames = \
                list(set([line[17:20].strip() for line in self.pdb_content
                          if line.startswith('HETATM') and line[21:22] == chain
                          and not line[17:20].strip() == 'HOH']))
            core_constraints = \
                ligand_core_constraints if ligand_resname in resnames else []
            d[chain] = {'Residues names': resnames,
                        'Core constraints': core_constraints}

        molecules = [self._extract_molecules_from_chain(
            chain=chain_id,
            rotamer_resolution=rotamer_resolution,
            exclude_terminal_rotamers=exclude_terminal_rotamers,
            allow_undefined_stereo=allow_undefined_stereo,
            core_constraints=d[chain_id]['Core constraints'])
            for chain_id in chain_ids]

        return sum(molecules, [])

    def get_molecules_from_chain(self, selected_chain, rotamer_resolution=30,
                                 exclude_terminal_rotamers=True,
                                 allow_undefined_stereo=False,
                                 core_constraints=[]):
        """
        It returns all hetero molecule defined in a specific chain from the
        PDB file. It handles the possible errors when selecting the chain
        for a PDB.

        Parameters
        ----------
        selected_chain : str
            Chain Id.
        rotamer_resolution : float
            The resolution in degrees to discretize the rotamer's
            conformational space. Default is 30
        exclude_terminal_rotamers : bool
            Whether to exclude terminal rotamers when generating the
            rotamers library  or not
        allow_undefined_stereo : bool
            Whether to allow a molecule with undefined stereochemistry
            to be defined or try to assign the stereochemistry and
            raise a complaint if not possible. Default is False

        Returns
        -------
        molecule : list[peleffy.topology.Molecule]
            The list of peleffy's Molecule object corresponding to the
            selected chain
        """

        chain_ids = set([line[21:22] for line in self.pdb_content
                         if line.startswith('HETATM')
                         and not line[17:20].strip() == 'HOH'])

        all_chain_ids = set([line[21:22] for line in self.pdb_content
                             if line.startswith('ATOM')
                             or line.startswith('HETATM')])

        if not selected_chain in all_chain_ids:
            raise ValueError('The selected chain {} '.format(selected_chain)
                             + 'is not a valid chain for this PDB. Available '
                             + 'chains to select are: {}'.format(chain_ids))

        if not selected_chain in chain_ids and selected_chain in all_chain_ids:
            raise ValueError('The selected chain {} '.format(selected_chain)
                             + 'is not an hetero molecule. Peleffy '
                             + 'is only compatible with hetero atoms.')

        molecules = self._extract_molecules_from_chain(
            chain=selected_chain,
            rotamer_resolution=rotamer_resolution,
            exclude_terminal_rotamers=exclude_terminal_rotamers,
            allow_undefined_stereo=allow_undefined_stereo,
            core_constraints=core_constraints)

        return molecules



"""
This module handles the selection of the right force field or parameter
calculators, given a context.
"""


class ForceFieldSelector(object):
    """
    It defines a force field selector.
    """
    _FF_TYPES = {'OPLS2005': ('OPLS2005'),
                 'OpenFF': ('openff_unconstrained-1.3.0.offxml',
                            'openff_unconstrained-1.2.1.offxml',
                            'openff_unconstrained-1.2.0.offxml',
                            'openff_unconstrained-1.1.1.offxml',
                            'openff_unconstrained-1.1.0.offxml',
                            'openff_unconstrained-1.0.1.offxml',
                            'openff_unconstrained-1.0.0.offxml')}

    def get_by_name(self, forcefield_name):
        """
        Given a forcefield name, it returns the corresponding
        force field class.

        Parameters
        ----------
        forcefield_name : str
            The name of the requested forcefield

        Returns
        -------
        forcefield : a peleffy.forcefield.forcefield.ForceField
            The force field that corresponds to the supplied name

        Raises
        ------
        ValueError
            If the supplied forcefield_name is unknown

        Examples
        --------

        Use the force field selector to select a force field by name

        >>> from peleffy.forcefield import ForceFieldSelector

        >>> selector = ForceFieldSelector()
        >>> openff = selector.get_by_name('openff_unconstrained-1.2.1.offxml')

        """
        from peleffy.utils import Logger

        log = Logger()
        log.info(' - Loading \'{}\''.format(forcefield_name))

        from .forcefield import (OpenForceField, OPLS2005ForceField)

        if forcefield_name.upper() in self._FF_TYPES['OPLS2005']:
            return OPLS2005ForceField()
        elif forcefield_name in self._FF_TYPES['OpenFF']:
            return OpenForceField(forcefield_name=forcefield_name)
        else:
            raise ValueError('Invalid force field name')

    def get_list(self):
        """
        It returns the complete list of available force fields.

        Returns
        -------
        forcefields : dict
            The complete list of available force fields grouped by
            force field type
        """
        return self._FF_TYPES


class ChargeCalculatorSelector(object):
    """
    It defines a charge calculator selector.
    """

    from peleffy.forcefield.calculators import (OPLSChargeCalculator,
                                                Am1bccCalculator,
                                                GasteigerCalculator,
                                                DummyChargeCalculator)

    _AVAILABLE_TYPES = {'opls2005': OPLSChargeCalculator,
                        'am1bcc': Am1bccCalculator,
                        'gasteiger': GasteigerCalculator,
                        'dummy': DummyChargeCalculator
                        }

    def get_by_name(self, charge_method, molecule):
        """
        Given a charge method name, it returns the corresponding
        charge calculator class.

        Parameters
        ----------
        charge_method : str
            The name of the requested charge calculator
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object whose partial charges will
            be calculated

        Returns
        -------
        charge_calculator : a PartialChargesCalculator
            The charge calculation method that will be employed to
            calculate partial charges

        Raises
        ------
        ValueError
            If the requested charge method is unknown
        """

        if charge_method.lower() not in self._AVAILABLE_TYPES:
            raise ValueError('Charge method \'{}\' '.format(charge_method)
                             + 'is unknown')

        charge_method = self._AVAILABLE_TYPES[charge_method.lower()]

        return charge_method(molecule)

    def get_list(self):
        """
        It returns the complete list of available force fields.

        Returns
        -------
        forcefields : dict
            The complete list of available force fields grouped by
            force field type
        """
        return self._FF_TYPES



"""
This module contains classes that handle force field representations.
"""


__all__ = ["OpenForceField", "OPLS2005ForceField",
           "OpenFFOPLS2005ForceField"]


from peleffy.forcefield.selectors import ChargeCalculatorSelector


class _BaseForceField(object):
    """
    It is the force field base class.
    """

    _type = ''
    _default_charge_method = 'am1bcc'

    def __init__(self, forcefield_name):
        """
        It initializes the base force field class.

        Parameters
        ----------
        forcefield_name : str
            The name of the forcefield

        Examples
        --------

        Load a molecule and generate its parameters with an OpenFF
        force field

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule)

        Load a molecule and generate its parameters with an OpenFF
        force field and 'gasteiger' partial charges

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenForceField

        >>> openff = OpenForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = openff.parameterize(molecule,
                                             charge_method='gasteiger')
        """
        self._name = forcefield_name

    def _get_charge_calculator(self, charge_method, molecule):
        """
        Given a charge method name, it returns the corresponding
        charge calculator class.

        Parameters
        ----------
        charge_method : str
            The name of the requested charge calculator
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize

        Returns
        -------
        charge_calculator : a PartialChargesCalculator object
            The charge calculation method that will be employed to calculate
            partial charges
        """

        # If charge method is not supplied, use force field's default
        if charge_method is None:
            charge_method = self._default_charge_method

        selector = ChargeCalculatorSelector()
        return selector.get_by_name(charge_method, molecule)

    def parameterize(self, molecule, charge_method=None):
        """
        It parameterizes the supplied molecule.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize
        charge_method : str
            The name of the charge method to employ

        Returns
        -------
        parameters : a peleffy.forcefield.parameters.BaseParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field
        """

        # Assign parameters
        parameters = self._get_parameters(molecule)

        # Assign partial charges using the charge calculator object
        charge_calculator = self._get_charge_calculator(charge_method,
                                                        molecule)
        charge_calculator.assign_partial_charges(parameters)

        return parameters

    @property
    def type(self):
        """
        It returns the type of the force field.

        Returns
        -------
        type : str
            The type of the force field
        """
        return self._type

    @property
    def name(self):
        """
        It returns the name of the force field.

        Returns
        -------
        name : str
            The name of the force field
        """
        return self._name

    @property
    def parameters(self):
        """
        It returns the parameters obtained with the force field.

        Returns
        -------
        parameters : any class deriving from BaseParameterWrapper
            The parameters obtained with the force field
        """
        return self._parameters


class OpenForceField(_BaseForceField):
    """
    It defines an OpenFF force field.
    """

    _type = 'OpenFF'
    _default_charge_method = 'am1bcc'

    def _get_parameters(self, molecule):
        """
        It parameterizes the supplied molecule with the OpenFF toolkit.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize

        Returns
        -------
        parameters : an OpenForceFieldParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field
        """
        from peleffy.utils.toolkits import OpenForceFieldToolkitWrapper

        openforcefield_toolkit = OpenForceFieldToolkitWrapper()
        parameters = openforcefield_toolkit.get_parameters_from_forcefield(
            self.name, molecule)

        from peleffy.forcefield.parameters \
            import OpenForceFieldParameterWrapper

        return OpenForceFieldParameterWrapper.from_label_molecules(
            molecule, parameters, self.name)


class OPLS2005ForceField(_BaseForceField):
    """
    It defines an OPLS2005 force field.
    """

    _type = 'OPLS2005'
    _default_charge_method = 'opls2005'

    def __init__(self):
        """
        It initializes the OPLS2005 force field class.

        Examples
        --------

        Load a molecule and generate its parameters with the OPLS2005
        force field

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OPLS2005ForceField

        >>> opls2005 = OPLS2005ForceField()
        >>> parameters = opls2005.parameterize(molecule)

        """

        super().__init__(self._type)

    def _get_parameters(self, molecule):
        """
        It parameterizes the supplied molecule with Schrodinger toolkit.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize

        Returns
        -------
        parameters : an OpenForceFieldParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field
        """
        from peleffy.utils.toolkits import SchrodingerToolkitWrapper

        schrodinger_toolkit = SchrodingerToolkitWrapper()
        ffld_output = schrodinger_toolkit.run_ffld_server(molecule)

        from peleffy.forcefield.parameters \
            import OPLS2005ParameterWrapper

        return OPLS2005ParameterWrapper.from_ffld_output(molecule,
                                                         ffld_output)


class OpenFFOPLS2005ForceField(_BaseForceField):
    """
    It defines an hybrid force field from the combination of OpenFF and
    OPLS2005.
    """

    _type = 'OpenFF + OPLS2005'
    _selections = ('openff', 'opls2005')
    _default_charge_method = 'am1bcc'

    def __init__(self, forcefield_name):
        """
        It initializes the OpenFFOPLS2005ForceField class.

        Parameters
        ----------
        openff_name : str
            The name of the OpenFF forcefield to employ

        Examples
        --------

        Load a molecule and generate its parameters with an hybrid
        OpenFF-OPLS2005 force field

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenFFOPLS2005ForceField

        >>> hybridff = OpenFFOPLS2005ForceField('openff_unconstrained-1.2.1.offxml')
        >>> parameters = hybridff.parameterize(molecule)

        We can customized it by selecting which force field employ for each
        parameter type. For example, below we are parameterizing all
        force field elements with OPLS2005 except for the dihedrals

        >>> from peleffy.topology import Molecule

        >>> molecule = Molecule('molecule.pdb')

        >>> from peleffy.forcefield import OpenFFOPLS2005ForceField

        >>> hybridff = OpenFFOPLS2005ForceField('openff_unconstrained-1.2.1.offxml')

        >>> hybridff.set_nonbonding_parameters('OPLS2005')
        >>> hybridff.set_bond_parameters('OPLS2005')
        >>> hybridff.set_angle_parameters('OPLS2005')
        >>> hybridff.set_dihedral_parameters('OpenFF')

        >>> parameters = hybridff.parameterize(molecule)

        """
        self._openff = OpenForceField(forcefield_name)
        self._oplsff = OPLS2005ForceField()
        super().__init__(self._openff.name + ' + ' + self._oplsff.name)

        # Set default selections
        self._nonbonding = 'openff'
        self._bonds = 'openff'
        self._angles = 'openff'
        self._dihedrals = 'openff'

    def _check_selection(self, selection):
        """
        It performs a safety check on the parameter selection.

        Parameters
        ----------
        selection : str
            The force field to employ. One of ('OpenFF', 'OPLS2005')

        Raises
        ------
        ValueError if the selection string is unknown
        """
        if selection.lower() not in self._selections:
            raise ValueError('Invalid selection: \'{}\'. '.format(selection)
                             + 'Valid selections are '
                             + '{}'.format(self._selections))

    def set_nonbonding_parameters(self, selection):
        """
        It sets which force field to use to assign nonbonding parameters.
        Nonbonding parameters are: 'atom_types', 'sigmas', 'epsilons',
        'SGB_radii', 'vdW_radii', 'gammas', 'alphas', 'GBSA_radii' and
        'GBSA_scales'.

        Parameters
        ----------
        selection : str
            The force field to employ. One of ('OpenFF', 'OPLS2005')
        """
        self._check_selection(selection)

        self._nonbonding = selection.lower()

    def set_bond_parameters(self, selection):
        """
        It sets which force field to use to assign parameters to bonds.

        Parameters
        ----------
        selection : str
            The force field to employ. One of ('OpenFF', 'OPLS2005')
        """
        self._check_selection(selection)

        self._bonds = selection.lower()

    def set_angle_parameters(self, selection):
        """
        It sets which force field to use to assign parameters to angles.

        Parameters
        ----------
        selection : str
            The force field to employ. One of ('OpenFF', 'OPLS2005')
        """
        self._check_selection(selection)

        self._angles = selection.lower()

    def set_dihedral_parameters(self, selection):
        """
        It sets which force field to use to assign parameters to dihedrals.

        Parameters
        ----------
        selection : str
            The force field to employ. One of ('OpenFF', 'OPLS2005')
        """
        self._check_selection(selection)

        self._dihedrals = selection.lower()

    def _get_parameters(self, molecule):
        """
        It parameterizes the supplied molecule with both OpenFF and
        Schrodinger toolkits.

        Parameters
        ----------
        molecule : a peleffy.topology.Molecule
            The peleffy's Molecule object to parameterize

        Returns
        -------
        parameters : an OpenForceFieldParameterWrapper object
            The parameter wrapper containing the parameters generated
            with the current force field
        """
        from peleffy.forcefield.parameters \
            import OpenFFOPLS2005ParameterWrapper

        hybrid_parameters = OpenFFOPLS2005ParameterWrapper(
            forcefield_name='{} + {}'.format(self._openff.name,
                                             self._oplsff.name))

        openff_parameters = self._openff.parameterize(molecule,
                                                      charge_method='dummy')
        print(openff_parameters['sigmas'])
        oplsff_parameters = self._oplsff.parameterize(molecule)
        print(oplsff_parameters['sigmas'])

        if self._nonbonding == 'openff':
            hybrid_parameters['atom_names'] = openff_parameters['atom_names']
            hybrid_parameters['atom_types'] = openff_parameters['atom_types']
            hybrid_parameters['charges'] = openff_parameters['charges']
            hybrid_parameters['sigmas'] = openff_parameters['sigmas']
            hybrid_parameters['epsilons'] = openff_parameters['epsilons']
            hybrid_parameters['SGB_radii'] = openff_parameters['SGB_radii']
            hybrid_parameters['vdW_radii'] = openff_parameters['vdW_radii']
            hybrid_parameters['gammas'] = openff_parameters['gammas']
            hybrid_parameters['alphas'] = openff_parameters['alphas']
            hybrid_parameters['GBSA_radii'] = openff_parameters['GBSA_radii']
            hybrid_parameters['GBSA_scales'] = openff_parameters['GBSA_scales']
        elif self._nonbonding == 'opls2005':
            hybrid_parameters['atom_names'] = oplsff_parameters['atom_names']
            hybrid_parameters['atom_types'] = oplsff_parameters['atom_types']
            hybrid_parameters['charges'] = oplsff_parameters['charges']
            hybrid_parameters['sigmas'] = oplsff_parameters['sigmas']
            hybrid_parameters['epsilons'] = oplsff_parameters['epsilons']
            hybrid_parameters['SGB_radii'] = oplsff_parameters['SGB_radii']
            hybrid_parameters['vdW_radii'] = oplsff_parameters['vdW_radii']
            hybrid_parameters['gammas'] = oplsff_parameters['gammas']
            hybrid_parameters['alphas'] = oplsff_parameters['alphas']
            hybrid_parameters['GBSA_radii'] = oplsff_parameters['GBSA_radii']
            hybrid_parameters['GBSA_scales'] = oplsff_parameters['GBSA_scales']
        else:
            raise ValueError('Invalid selection: '
                             + '\'{}\'. '.format(self._nonbonding)
                             + 'Valid selections are '
                             + '{}'.format(self._selections))

        if self._bonds == 'openff':
            hybrid_parameters['bonds'] = openff_parameters['bonds']
        elif self._bonds == 'opls2005':
            hybrid_parameters['bonds'] = oplsff_parameters['bonds']
        else:
            raise ValueError('Invalid selection: '
                             + '\'{}\'. '.format(self._bonds)
                             + 'Valid selections are '
                             + '{}'.format(self._selections))

        if self._angles == 'openff':
            hybrid_parameters['angles'] = openff_parameters['angles']
        elif self._angles == 'opls2005':
            hybrid_parameters['angles'] = oplsff_parameters['angles']
        else:
            raise ValueError('Invalid selection: '
                             + '\'{}\'. '.format(self._angles)
                             + 'Valid selections are '
                             + '{}'.format(self._selections))

        if self._dihedrals == 'openff':
            hybrid_parameters['propers'] = openff_parameters['propers']
            hybrid_parameters['impropers'] = openff_parameters['impropers']
        elif self._dihedrals == 'opls2005':
            hybrid_parameters['propers'] = oplsff_parameters['propers']
            hybrid_parameters['impropers'] = oplsff_parameters['impropers']
        else:
            raise ValueError('Invalid selection: '
                             + '\'{}\'. '.format(self._dihedrals)
                             + 'Valid selections are '
                             + '{}'.format(self._selections))

        return hybrid_parameters
